/* 
 * Filters go here 
 */

#include <stdio.h>
#include <math.h>
#include "moused.h"
#include "filter_macros.h"

/* rodent comes from moused.c */
extern rodent_t rodent;

/**
 * Virtual Scrolling
 */
static int scroll_state;
static int scroll_distance;

#define SCROLL_DISTANCE_THRESHOLD 3

#define SCROLL_READY 0
#define SCROLL_PREPARED 1
#define SCROLL_ING 2
#define SCROLL_TIMEOUT 3

/* Virtual Scrolling filter */
static void filter_virtual_scroll(mouse_info_t *delta, int noop) {
	if (noop == 1) { /* timeout */
		warnx("timeout!");
		if (scroll_state == SCROLL_PREPARED) {
			RESTORE_STATE();
			scroll_state = SCROLL_TIMEOUT;
		}
	} else {
		if (GET_BUTTON_STATE(2) > 0) {
			if (scroll_state == SCROLL_TIMEOUT)
				return;

			if (scroll_state == SCROLL_READY)
				scroll_state = SCROLL_PREPARED;

			if (GET_MOUSE_X() != 0 || GET_MOUSE_Y() != 0) {
				if (scroll_state == SCROLL_PREPARED)
					scroll_state = SCROLL_ING;

				scroll_distance += GET_MOUSE_Y();

				if (scroll_distance > rodent.virtual_scroll_threshold) {
					SET_MOUSE_Z(1);
					scroll_distance = 0;
				} else if (scroll_distance < 0 - rodent.virtual_scroll_threshold) {
					SET_MOUSE_Z(-1);
					scroll_distance = 0;
				}

				/* Don't move! */
				SET_MOUSE_X(0);
				SET_MOUSE_Y(0);
			}

			/* Don't click middle */
			CLEAR_BUTTON_STATE(2);

			if (SCROLL_PREPARED == scroll_state)
				NEED_TIMEOUT(100);
		} else {
			if (SCROLL_PREPARED == scroll_state) {
				SET_BUTTON_STATE(2);
				RAW_UPDATE();
				CLEAR_BUTTON_STATE(2);
			}
			scroll_state = SCROLL_READY;
		}
	}
}

/* Linear Acceleration filter */
static void filter_linear_acceleration(mouse_info_t *delta, int noop) {
	SET_MOUSE_X(GET_MOUSE_X() * rodent.accelx);
	SET_MOUSE_Y(GET_MOUSE_Y() * rodent.accely);
}

/* Exponential Acceleration filter */

static struct expaccel {
	double x;
	double y;
	struct timeval *tv;
} expaccel = { .tv = NULL };

static double expscale(double val) {
	int neg = 0;
	if (val < 0) {
		neg = 1;
		val *= -1.0;
	}

	//val = pow(val, 2);
	val /= 10.0;
	val = log1p(val);

	if (neg)
		val *= -1.0;

	return val;
}
static void filter_exponential_acceleration(mouse_info_t *delta, int noop) {
	/* Multiply mouse movement based on instantaneous speed? */
	struct timeval tv;
	double timedelta;
	double delta_x, delta_y;

	gettimeofday(&tv, NULL);

	if (expaccel.tv == NULL) {
		expaccel.tv = malloc(sizeof(struct timeval));
		memcpy(expaccel.tv, &tv, sizeof(struct timeval));
		return;
	}

	/* Speed = timedelta / distancedelta */
	timedelta = (tv.tv_sec - expaccel.tv->tv_sec) + \
					((tv.tv_usec - expaccel.tv->tv_usec + 0.0) / (1000.0 * 1000.0));

	delta_x = (GET_MOUSE_X() + 0.0) / timedelta;
	delta_y = (GET_MOUSE_Y() + 0.0) / timedelta;

	//expaccel.x += pow(delta_x,.2);
	//expaccel.y += pow(delta_y,.2);

	//printf("delta1 [%08.08f]: %f,%f\n", timedelta, delta_x, delta_y);
	//printf("delta2 [%08.08f]: %f,%f\n", timedelta, expscale(delta_x), expscale(delta_y));
	printf("\n");
	//printf("log [%f]: %f,%f\n", timedelta, pow(delta_x,.5), pow(delta_y,.5));

	if ((expaccel.x >= 1) || (expaccel.y <= -1)) {
		//SET_MOUSE_X((char)floor(expaccel.x));
		expaccel.x -= floor(expaccel.x);
	}
	//else
		//SET_MOUSE_X(0);

	if ((expaccel.y >= 1) || (expaccel.y <= -1)) {
		//SET_MOUSE_Y((char)floor(expaccel.y));
		expaccel.y -= floor(expaccel.y);
	}
	//else
		//SET_MOUSE_Y(0);

	//printf("%d,%d\n", GET_MOUSE_X(), GET_MOUSE_Y());

	memcpy(expaccel.tv, &tv, sizeof(struct timeval));
}

/**
 * Emulate 3 Button filter
 */

static int emulate3_state;
#define EMULATE_WAITING 1
#define EMULATE_ING 2

static void filter_emulate_3_button(mouse_info_t *delta, int noop) {
	if (noop == 1) {
		emulate3_state = 0;
		RESTORE_STATE();
	} else {
		/* Check if we're holding left and right buttons */
		if (GET_BUTTON_STATE(1) > 0 && GET_BUTTON_STATE(3) > 0) {
			if (EMULATE_WAITING == emulate3_state)
				emulate3_state = EMULATE_ING;
			if (EMULATE_ING == emulate3_state)
				SET_BUTTON_STATE(2);
		} else if (GET_BUTTON_STATE(1) > 0 || GET_BUTTON_STATE(3)) {
			emulate3_state = EMULATE_WAITING;
			NEED_TIMEOUT(50);
		} else {
			if (EMULATE_ING == emulate3_state) {
			}
			emulate3_state = 0;
		}
		CLEAR_BUTTON_STATE(1);
		CLEAR_BUTTON_STATE(3);
	}
}

/**
 * Chord Middle
 */

FILTER(chord_middle) {
	/* XXX: Chord middle.... */
}

void filter(mouse_info_t *delta, int noop) {
	if (rodent.flags & FLAG_ACCELERATION)
		//filter_linear_acceleration(delta, noop);
		filter_exponential_acceleration(delta, noop);

	if (rodent.flags & FLAG_EMULATE_3_BUTTON)
		filter_emulate_3_button(delta, noop);

	if (rodent.flags & FLAG_CHORD_MIDDLE)
		filter_chord_middle(delta, noop);

	if (rodent.flags & FLAG_VIRTUAL_SCROLL)
		filter_virtual_scroll(delta, noop);
}
