/*
 * Synaptics touchpad
 *
 */

#include <sys/types.h>
#include <sys/mouse.h>
#include <sys/consio.h>
#include <sys/time.h>

#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <err.h>
#include <errno.h>
#include "../../moused.h"
#include "../../lib/ps2/ps2.h"

#define MOUSE_BUTTON(x) (1 << (x - 1))
#define SYNAPTICS_X (((packet[3] & 0x10) << 8) | \
                     ((packet[1] & 0x0F) << 8) | \
                     packet[4] & 0xFF)

#define SYNAPTICS_Y (((packet[3] & 0x20) << 7) | \
                     ((packet[1] & 0xF0) << 4) | \
                     packet[5] & 0xFF)

#define tv2int(x) (((x).tv_sec * 1000000) + (x).tv_usec)
#define TAP_TIMEOUT 125000
#define LOW_SPEED_THRESHOLD 20
#define SQUELCH_LEVEL 3
#define MIN_MOVEMENT 2

/* "pressure" 20-30 still shows up even if I'm not touching the pad */
#define TOUCH_THRESHOLD 35

/* "tap" pressure minimum */
#define TAP_THRESHOLD 40

typedef struct synaptics {
	int w; /* palm/finger detection */
	int x; /* absolute x coordinate */
	int y; /* absolute y coordinate */
	int z; /* z pressure (capacitance measure) */
	int buttons; /* bitmap of buttons */

	int guestbuttons;
	int touchbuttons;
	int clickdrag;

	/* Last packet's coordinates */
	int finger;
	int oldx;
	int oldy;
	struct timeval lastfingertime;
	int tap;
	int zmax;
	int drag;

	/* Rolling averages */
	int xaverage;
	int yaverage;
	int squelch;
} synaptics_t;

/* From libkern.h */
static __inline int imax(int a, int b) { return (a > b ? a : b); }

static void printbits(char val) {
	int i;
	for (i = 7; i >= 0; i--) {
		printf("%d", (val >> i) & 1);
	}
}

static void printbits_int(int val) {
	int i;
	for (i = sizeof(val) * 8 - 2; i >= 0; i--) {
		printf("%d", (val >> i) & 1);
	}
}

static void activity(rodent_t *rodent, char *packet);
static void (*logmsg)(int, int, const char *, ...) = NULL;
static void smoothdata(mouse_info_t *delta);
void guestdata(char *packet, rodent_t *rodent);
void touchpaddata(char *packet, rodent_t *rodent);
static int timecmplt(struct timeval *tv, int timeout);

static synaptics_t synaptics;

MOUSED_INIT_FUNC {
	logmsg = rodent->logmsg;
}

MOUSED_PROBE_FUNC {
	char status[3];

	/* XXX: Shouldn't moused.c do the open? */
	rodent->mfd = open(rodent->device, O_RDWR);
	if (rodent->mfd == -1)
		logfatal(1, "Unable to open %s", rodent->device);

	ps2_openfd(rodent->mfd);
	ps2_reset();

	ps2_disable_stream();

	/* Synaptics Probe, as per documentation */
	ps2_ext_command(0);
	ps2_get_status(status);

	if (status[1] != 0x47) {
		warnx("This is not a Synaptics device");
		return MODULE_PROBE_FAIL;
	}

	/* Re-enable streaming */
	ps2_enable_stream();

	if (rodent->verbose)
		warnx("Synaptics Touchpad version %d.%d", status[2] & 0x0F, status[0]);

	close(rodent->mfd);

	return MODULE_PROBE_SUCCESS;
}

MOUSED_HANDLER_INIT_FUNC {
	char status[3];
	if (rodent->verbose > 1)
		warnx("synaptics init");

	rodent->mfd = open(rodent->device, O_RDWR);
	if (rodent->mfd == -1)
		logfatal(1, "Unable to open %s", rodent->device);

	ps2_openfd(rodent->mfd);
	ps2_reset();

	rodent->mode.packetsize = 6;
	if (NULL == rodent->packet) {
		rodent->packet = malloc(rodent->mode.packetsize);
	}

	ps2_set_packetsize(rodent->mode.packetsize);

	/*
	 * Put the touchpad in "absolute mode"
	 * This is done by encoding an 8-bit argument in 4 set_resolution commands
	 * and then sending an Set Sample Rate 20 command
	 */

	/* First, disable device reporting */
	ps2_disable_stream();

	/* Enable Absolute mode, high packet rate, and Wmode ( 0xC1 ) */
	ps2_ext_command(0xc1);
	ps2_set_sample_rate(20); /* 0x14 */

	/* Request 'Synaptics Mode' status */
	ps2_ext_command(1);
	ps2_get_status(status);

	ps2_enable_stream();

	/* Set the syncmask of our protocol header (first byte) */
	ps2_set_syncmask(0xc8);
}

MOUSED_HANDLER_FUNC {
	int bytes;

	/* XXX: Should this be done in libps2?
	 * ie; bytes = ps2_read_packet(rodent->packet)
	 */
	bytes = read(rodent->mfd, rodent->packet, rodent->mode.packetsize);

	if (bytes < 0)
		return;

	if (bytes == rodent->mode.packetsize)
		activity(rodent, rodent->packet);
	else
		warnx("Invalid packet size: %d", bytes);
}

static void activity(rodent_t *rodent, char *packet) {
	mouse_info_t delta;

	delta.operation = MOUSE_ACTION;
	delta.u.data.buttons = 0;
	delta.u.data.x = delta.u.data.y = delta.u.data.z = 0;

	synaptics.buttons = synaptics.w = synaptics.x = synaptics.y = synaptics.z = 0;

	/* Finger "type" value */
	synaptics.w = ((packet[0] & 0x30) >> 2) |
	         ((packet[0] & 0x04) >> 1) |
	         ((packet[3] & 0x04) >> 2);

	/* Pressure/Contact level */
	synaptics.z = packet[2] & 0xff;

	/* Guest devices (trackpoint, for example) show up with 'w' of 3 */
	if (synaptics.w == 3)
		guestdata(packet, rodent);
	else
		touchpaddata(packet, rodent);

	if (synaptics.drag == 1) /* Touchpad dragging */
		smoothdata(&delta);
	else if (synaptics.w == 3) { /* Guest device */
		delta.u.data.x = synaptics.x;
		delta.u.data.y = synaptics.y;
	}

	/* We keep state on both guest and touchpad mouse buttons */
	delta.u.data.buttons = synaptics.guestbuttons | synaptics.touchbuttons;
	/* XXX: y needs to be negated? */
	delta.u.data.y = -delta.u.data.y;

	synaptics.oldx = synaptics.x;
	synaptics.oldy = synaptics.y;
	synaptics.zmax = imax(synaptics.z, synaptics.zmax);

	if (rodent->verbose > 1)
		if (delta.u.data.x != 0 || delta.u.data.y != 0 || delta.u.data.buttons != 0
			 || rodent->verbose > 2)
			printf("Update: (%d,%d) %x\n", delta.u.data.x, delta.u.data.y, delta.u.data.buttons);

	rodent->update(&delta);
}

static void smoothdata(mouse_info_t *delta) {
	int xa, ya; /* X and Y averages */
	int x, y;
	int sensitivity, xsensitivity, ysensitivity;

	if (synaptics.xaverage == 0)
		synaptics.xaverage = synaptics.x;
	if (synaptics.yaverage == 0)
		synaptics.yaverage = synaptics.y;

	xa = synaptics.xaverage;
	ya = synaptics.yaverage;

	synaptics.xaverage = (xa + synaptics.x) / 2;
	synaptics.yaverage = (ya + synaptics.y) / 2;

	xsensitivity = (synaptics.xaverage - xa);
	if (xsensitivity < 0)
		xsensitivity = -xsensitivity;
	ysensitivity = (synaptics.yaverage - ya);
	if (ysensitivity < 0)
		ysensitivity = -ysensitivity;

	sensitivity = (xsensitivity & ysensitivity) + 1;

	if ((synaptics.x > LOW_SPEED_THRESHOLD || synaptics.x < -LOW_SPEED_THRESHOLD)
		 || (synaptics.y > LOW_SPEED_THRESHOLD || synaptics.y < -LOW_SPEED_THRESHOLD)) {
		synaptics.x = (synaptics.x + synaptics.oldx * 3) / 4;
		synaptics.y = (synaptics.y + synaptics.oldy * 3) / 4;

		delta->u.data.x = (synaptics.x - synaptics.oldx) * 10 / 85;
		delta->u.data.y = (synaptics.y - synaptics.oldy) * 10 / 85;
	} else {
		if (synaptics.x < -MIN_MOVEMENT)
			delta->u.data.x = -1;
		else if (synaptics.x > MIN_MOVEMENT)
			delta->u.data.x = 1;
		else
			delta->u.data.x = 0;

		if (synaptics.y < -MIN_MOVEMENT)
			delta->u.data.y = -1;
		else if (synaptics.y > MIN_MOVEMENT)
			delta->u.data.y = 1;
		else
			delta->u.data.y = 0;
	}

	synaptics.squelch += sensitivity;

	if (synaptics.squelch < SQUELCH_LEVEL) {
		delta->u.data.x = 0;
		delta->u.data.y = 0;
	} else 
		synaptics.squelch = 0;

	//printf("(movement) (%d, %d) (%02x)\n", (synaptics.x - synaptics.oldx), synaptics.y - synaptics.oldy, synaptics.buttons);
}

void guestdata(char *packet, rodent_t *rodent) {
	synaptics.x = (packet[1] & 0x10) ? packet[4] : packet[4];
	synaptics.y = (packet[1] & 0x20) ? packet[5]  : packet[5];
	synaptics.z = 0;

	synaptics.guestbuttons = 0;
	if (packet[1] & 0x01)
		synaptics.guestbuttons |= MOUSE_BUTTON(1);
	if (packet[1] & 0x02)
		synaptics.guestbuttons |= MOUSE_BUTTON(3);
	if (packet[1] & 0x04)
		synaptics.guestbuttons |= MOUSE_BUTTON(2);

	//printf("(nipple) (%d, %d) W:%u Z:%u (%02x)\n", synaptics.x, synaptics.y, synaptics.w, synaptics.z, synaptics.buttons);
}

void touchpaddata(char *packet, rodent_t *rodent) {
	synaptics.x = SYNAPTICS_X;
	synaptics.y = SYNAPTICS_Y;

	synaptics.touchbuttons = 0;

	if (packet[0] & 0x01)
		synaptics.touchbuttons |= MOUSE_BUTTON(1);
	if (packet[0] & 0x02)
		synaptics.touchbuttons |= MOUSE_BUTTON(3);

	/* Single fingers are 4 <= w <= 8 */
	if ((synaptics.w >= 4 && synaptics.w <= 8) 
		 && (synaptics.z >= TOUCH_THRESHOLD && synaptics.z < 200)) {

		if (rodent->verbose)
			printf("Single finger\n");

		if (synaptics.tap > 0) {
			if (timecmplt(&synaptics.lastfingertime, TAP_TIMEOUT))
				synaptics.clickdrag = MOUSE_BUTTON(1);

			if (synaptics.clickdrag)
				synaptics.touchbuttons |= synaptics.clickdrag;
		}

		if (synaptics.finger == 0) {
			synaptics.finger = 1;
			gettimeofday(&synaptics.lastfingertime, NULL);
		} else {
			synaptics.drag = 1;
		}

		//printf("(touchpad) (%d, %d) W:%u Z:%u (%02x)\n", synaptics.x, synaptics.y, synaptics.w, synaptics.z, synaptics.buttons);
	}

	/* w == 0 indicates two fingers */
	else if (synaptics.w == 0 && synaptics.z > TOUCH_THRESHOLD) {

		if (rodent->verbose)
			printf("Two fingers\n");

		if (synaptics.finger == 0) {
			synaptics.finger = 2;
			gettimeofday(&synaptics.lastfingertime, NULL);
		} else {
			synaptics.drag = 1;
		}

		/* XXX: Only send right mouse if this is the first set of fingers down? */
		if (synaptics.finger == 2)
			synaptics.touchbuttons |= MOUSE_BUTTON(3);

		//printf("(twofinger) (%d, %d) W:%u Z:%u (%02x)\n", synaptics.x, synaptics.y, synaptics.w, synaptics.z, synaptics.buttons);
	}

	/* w == 1 indicates three or more fingers */
	else if (synaptics.w == 1 && synaptics.z > TOUCH_THRESHOLD) {

		if (rodent->verbose)
			printf("Three+ fingers\n");

		/* It's difficult to have 3-finger detected first before 2finger */
		if (synaptics.finger == 0 || synaptics.finger == 2) {
			synaptics.finger = 3;
			gettimeofday(&synaptics.lastfingertime, NULL);
		} else {
			synaptics.drag = 1;
		}

		/* XXX: Only send right mouse if this is the first set of fingers down? */
		if (synaptics.finger == 3)
			synaptics.touchbuttons |= MOUSE_BUTTON(2);

		//printf("(threeplus) (%d, %d) W:%u Z:%u (%02x)\n", synaptics.x, synaptics.y, synaptics.w, synaptics.z, synaptics.buttons);
	}
	/* Unknown W: value? */
	else {
		/* equate finger with button values */
		if (synaptics.z >= TOUCH_THRESHOLD) {
			printf("Unkown: W:%u Z:%u (%02x)\n", synaptics.w, synaptics.z, synaptics.touchbuttons);
		}
		if (synaptics.finger) {
			/* Finger was down previously, now it's not */

			/* If the finger was down less than TAP_TIMEOUT, then consider this a tap */
			if ((synaptics.zmax >= TAP_THRESHOLD) &&
				 timecmplt(&synaptics.lastfingertime, TAP_TIMEOUT)) {
				/* XXX: Some sort of special tap handling should go here.
				 * Handle coordinate-based taps for special clicks */
				printf("Tap!\n");
				synaptics.tap++;
			} else {
				if (rodent->verbose)
					printf("no tap detected\n");
				synaptics.tap = 0;
			}
			gettimeofday(&synaptics.lastfingertime, NULL);

			if (synaptics.tap) {
				printf("Clicking\n");
				if (synaptics.w == 0)
					synaptics.touchbuttons |= MOUSE_BUTTON(3);
				else if (synaptics.w == 1)
					synaptics.touchbuttons |= MOUSE_BUTTON(2);
				else if (synaptics.w >= 4 && synaptics.w <= 8)
					synaptics.touchbuttons |= MOUSE_BUTTON(1);
			}

			synaptics.clickdrag = 0;
			synaptics.finger = 0;
			synaptics.drag = 0;
			synaptics.oldx = synaptics.oldy = 0;

		}
	}

	if (rodent->verbose > 2)
		printf("(%s) W:%d Z:%d (%d, %d) %02x\n", (synaptics.w == 3) ? "nipple" : "finger", synaptics.w, synaptics.z, synaptics.x, synaptics.y, synaptics.touchbuttons);
}

static int timecmplt(struct timeval *tv, int timeout) {
	struct timeval cur;

	gettimeofday(&cur, NULL);
	return (tv2int(cur) - tv2int(*tv) < timeout);
}
