/* 
 * new moused(8). A flexible plugin-oriented framework 
 * for userland mouse drivers.
 *
 * (jls@freebsd.org)
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/select.h>
#include <sys/types.h>

/* sysmouse stuff */
#include <sys/consio.h>
#include <sys/mouse.h>

#include <dlfcn.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>

#include "moused.h"

static int nodaemon = 0;

static void usage();
static void logmsg(int log_pri, int errnum, const char *fmt, ...);
static int moused(int argc, char **argv);

/* Try to load a module into memory (dlopen, etc) */
static int loadmodule(char *path);

/* Modules will call this function when the mouse moves/changes */
static int update(mouse_info_t *delta);

/* Filter the input somehow? (comes from filters.c) */
extern void filter(mouse_info_t *delta, int noop);

static void packethandler_main();
void dumpstate(mouse_info_t *delta);

#define MOUSED_OPTIONS "d:m:f3E:VU:a:vh"

rodent_t rodent = {
	.device = NULL,
	.modulename = "auto", /* XXX: Should we default to autodetect? */
	.configfile = "/etc/moused.conf",
	.cfd = -1,
	.mfd = -1,

	.flags = 0,
	.packet = NULL,
	.emulate3_timeout = 0,
	.virtual_scroll_threshold = 3,
	.accelx = 1,
	.accely = 1,

	.update = update, /* Modules call moused's update(...) to talk to sysmouse */
	.logmsg = logmsg, /* Modules call moused's logmsg(...) log stuff */

	.init = NULL,
	.probe = NULL,
	.run = NULL,

	.handler = NULL,
};

int
main(int argc, char **argv)
{
	int c;
	int i;

	while (-1 != (c = getopt(argc, argv, MOUSED_OPTIONS))) {
		switch (c) {
		case 'f':
			nodaemon = 1;
			break;
		case '3':
			rodent.flags |= FLAG_EMULATE_3_BUTTON;
			break;
		case 'E':
			rodent.emulate3_timeout = atoi(optarg);
			break;
		case 'U':
			rodent.virtual_scroll_threshold = atoi(optarg);
			break;
		case 'V':
			if (rodent.verbose)
				warnx("enabling virtual scroll");
			rodent.flags |= FLAG_VIRTUAL_SCROLL;
			break;
		case 'a':
			rodent.flags |= FLAG_ACCELERATION;
			i = sscanf(optarg, "%f,%f",
			    &rodent.accelx, &rodent.accely);
			if (i == 0) {
				warnx("invalid acceleration argument '%s'",
				    optarg);
				//usage();
				rodent.accelx = rodent.accely = 1;
			}
			if (i == 1)
				rodent.accely = rodent.accelx;

			break;
		case 'd':
			rodent.device = optarg;
			break;
		case 'm':
			rodent.modulename = optarg;
			break;
		case 'v':
			rodent.verbose++;
			break;

		case 'h':
			usage();
			exit(0);
		default:
			usage();
			exit(1);
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if (rodent.device == NULL)
		rodent.device = "/dev/psm0";

	moused(argc, argv);

	return (0);
}

#define LIBRARY_PREFIX "moused_"

int
moused(int argc, char **argv)
{
	struct mouse_info mouse;
	char path[256];
	DIR *libdir;
	struct dirent *dp;

	if ((rodent.cfd = open("/dev/consolectl", O_RDWR, 0)) == -1)
		logfatal(1, "Unable to open /dev/consolectl");

	if (strcasecmp(rodent.modulename, "auto") == 0) {
		/* Look for "moused_*.so" files */
		libdir = opendir("/usr/lib");
		while ((dp = readdir(libdir)) != NULL) {
			if ((strncmp(dp->d_name, LIBRARY_PREFIX, strlen(LIBRARY_PREFIX)) == 0)
				 && (strcmp(dp->d_name + strlen(dp->d_name) - 3,".so") == 0)) {
				/* Try this module */
				if (rodent.verbose)
					warnx("Trying module: %s\n", dp->d_name);
				snprintf(path, 256, "/usr/lib/%s", dp->d_name);
				if (loadmodule(path) == 0)
					logfatal(1, "Unable open module '%s'.", dp->d_name);
				if (rodent.init != NULL)
					rodent.init(&rodent, argc, argv);
				if (rodent.probe == NULL)
					logfatal(1, "Module '%s' has no PROBE function?!",
								dp->d_name);

				if (rodent.probe(&rodent) != MODULE_PROBE_SUCCESS) {
					if (rodent.verbose)
						warnx("%s probe: FAILED", dp->d_name);
				} else {
					rodent.modulename = dp->d_name;
					if (rodent.verbose)
						warnx("%s probe: SUCCESS", dp->d_name);
					break;
				}
			}
		}
		closedir(libdir);
	} else {
		/* Load a specific module */
		snprintf(path, 256, "/usr/lib/moused_%s.so",
			 rodent.modulename, rodent.modulename);
		if (rodent.verbose)
			printf("Module path: %s\n", path);
		/* Try to load the module from system path */
		if (loadmodule(path) == 0) {
			logfatal(1, "Unable open module '%s'.", rodent.modulename);
		}
	}


	if (rodent.init != NULL) {
		/* Reset getopt stuffs (do it in case init()
		 * likes to use getopt)
		 */
		optreset = 1;
		optind = 0;
		rodent.init(&rodent, argc, argv);
	}

	if (rodent.probe == NULL)
		logfatal(1, "Module '%s' has no PROBE function?!",
			 rodent.modulename);

	/* Right module for this mouse device? */
	if (rodent.probe(&rodent) != MODULE_PROBE_SUCCESS) {
		warnx("Module probe failed: Mouse does not appear to be of type '%s'", rodent.modulename);
		exit(1);
	}

	if (nodaemon == 0 && daemon(0, 0) == -1)
			logfatal(1, "failed to become a daemon");

	setproctitle("[%s on %s]", rodent.modulename, rodent.device);

	if (rodent.handler != NULL)
		packethandler_main();
	else 
		logfatal(1, "Module '%s' has no "
			"RUN function?!",
			 rodent.modulename);

	/* If we _EVER_ get here, something hs wrong */
	warnx("Arrived at unexpected location "
		"(RUN or PACKETHANDLER method died?)");
}

static void
logmsg(int log_pri, int errnum, const char *fmt, ...)
{
	va_list ap;
	char buf[256];

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);

	if (errnum) {
		strlcat(buf, ": ", sizeof(buf));
		strlcat(buf, strerror(errnum), sizeof(buf));
	}

	/* Syslog if we're in the background? */
	warnx("%s", buf);
}

static int
loadmodule(char *path)
{
	void *handle;

	if ((handle = dlopen(path, RTLD_LAZY)) == NULL)
		return 0;

	rodent.init = dlsym(handle, "init");
	rodent.probe = dlsym(handle, "probe");
	rodent.run = dlsym(handle, "run");
	rodent.handler = dlsym(handle, "handler");
	rodent.handler_init = dlsym(handle, "handler_init");

	return 1;
}

static int
update(mouse_info_t *delta)
{
	mouse_info_t mydelta;

	/* Store state and copy delta into mydelta if this isn't a no-op */
	if (delta != NULL) {
		memcpy(&mydelta, delta, sizeof(mouse_info_t));
		memcpy(&(rodent.state), delta, sizeof(mouse_info_t));
	} else {
		memset(&mydelta, 0, sizeof(mouse_info_t));
	}

	/* XXX: Add calls to the filter stuffs here.
	 * Maybe filters need to go in separate .c files for readability?
	 * Link them in to the moused binary, then have filter() call them
	 * individually? Not sure yet */
	filter(&mydelta, (NULL == delta) ? 1 : 0);
	return ioctl(rodent.cfd, CONS_MOUSECTL, &mydelta);
}

static void
packethandler_main()
{
	fd_set fds;
	struct timeval timeout;

	timeout.tv_sec = -1; /* Infinite */
	timeout.tv_usec = 0; 

	if (rodent.handler_init != NULL)
		rodent.handler_init(&rodent);

	if (rodent.mfd == -1) {
		rodent.mfd = open(rodent.device, O_RDWR);
		if (rodent.mfd == -1) 
			logfatal(1, "Failed trying to open '%s'",
			    rodent.device);
	}

	FD_ZERO(&fds);
	FD_SET(rodent.mfd, &fds);
	
	for (;;) {
		int c;
		struct timeval *t;

		if ((rodent.sleeptime.tv_sec == 0) &&
		    (rodent.sleeptime.tv_usec == 0)) {
			t = NULL; /* Wait forever */
		} else {
			warnx("Using rodent sleeptime");
			t = &(rodent.sleeptime);
		}

		c = select(FD_SETSIZE, &fds, NULL, NULL, t);

		/* Reset sleeptime */
		rodent.sleeptime.tv_sec = rodent.sleeptime.tv_usec = 0;

		if (c < 0)
			logfatal(1, "Error with select call");
		else if (c > 0)
			rodent.handler(&rodent);
		else {
			warnx("SELECT TIMEOUT");
			rodent.update(NULL);
			if (rodent.timeout != NULL)
				rodent.timeout(&rodent);
		}

		FD_SET(rodent.mfd, &fds);
	}
}

void
dumpstate(mouse_info_t *delta)
{
	warnx("(%d,%d) z:%d b:%08x", delta->u.data.x, delta->u.data.y,
	    delta->u.data.z, delta->u.data.buttons);
}

void
usage() 
{
	printf("usage: moused [-d device] [-m modulename] [-fvh]\n");
	printf("              [-3 [-E timeout]] [-V [-U threshold]]\n");
	printf("              [-a accel]\n");
	printf(" -d device     specify the mouse device to use\n");
	printf(" -m module     specify the mouse module to use\n");
	printf(" -f            run in the foreground\n");
	printf(" -v            increase verbosity (multiple OK)\n");
	printf(" -V            enable 'virtual scrolling'\n");
	printf(" -U minmove    change virtual scroll movement threshold\n");
	printf(" -3            enable 'emulate 3 button'\n");
	printf(" -E timeout    set timeout for 'emulate 3 button' mode\n");
	printf(" -a x          set acceleration multiplier\n");
	printf(" -h            this help\n");
}
