/*-
 * Copyright (c) 1992, 1993 Erik Forsberg.
 * Copyright (c) 1996, 1997 Kazutaka YOKOTA.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * THIS SOFTWARE IS PROVIDED BY ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL I BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 *  Ported to 386bsd Oct 17, 1992
 *  Sandi Donno, Computer Science, University of Cape Town, South Africa
 *  Please send bug reports to sandi@cs.uct.ac.za
 *
 *  Thanks are also due to Rick Macklem, rick@snowhite.cis.uoguelph.ca -
 *  although I was only partially successful in getting the alpha release
 *  of his "driver for the Logitech and ATI Inport Bus mice for use with
 *  386bsd and the X386 port" to work with my Microsoft mouse, I nevertheless
 *  found his code to be an invaluable reference when porting this driver
 *  to 386bsd.
 *
 *  Further modifications for latest 386BSD+patchkit and port to NetBSD,
 *  Andrew Herbert <andrew@werple.apana.org.au> - 8 June 1993
 *
 *  Cloned from the Microsoft Bus Mouse driver, also by Erik Forsberg, by
 *  Andrew Herbert - 12 June 1993
 *
 *  Modified for PS/2 mouse by Charles Hannum <mycroft@ai.mit.edu>
 *  - 13 June 1993
 *
 *  Modified for PS/2 AUX mouse by Shoji Yuen <yuen@nuie.nagoya-u.ac.jp>
 *  - 24 October 1993
 *
 *  Hardware access routines and probe logic rewritten by
 *  Kazutaka Yokota <yokota@zodiac.mech.utsunomiya-u.ac.jp>
 *  - 3, 14, 22 October 1996.
 *  - 12 November 1996. IOCTLs and rearranging `psmread', `psmioctl'...
 *  - 14, 30 November 1996. Uses `kbdio.c'.
 *  - 13 December 1996. Uses queuing version of `kbdio.c'.
 *  - January/February 1997. Tweaked probe logic for 
 *    HiNote UltraII/Latitude/Armada laptops.
 *  - 30 July 1997. Added APM support.
 *  - 5 March 1997. Defined driver configuration flags (PSM_CONFIG_XXX). 
 *    Improved sync check logic.
 *    Vendor specific support routines.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD: src/sys/dev/atkbdc/psm.c,v 1.86 2005/06/10 20:56:37 marius Exp $");

#include "opt_isa.h"
#include "opt_psm.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/poll.h>
#include <sys/syslog.h>
#include <machine/bus.h>
#include <sys/rman.h>
#include <sys/selinfo.h>
#include <sys/sysctl.h>
#include <sys/time.h>
#include <sys/uio.h>

#include <sys/limits.h>
#include <sys/mouse.h>
#include <machine/resource.h>


#include <dev/atkbdc/atkbdcreg.h>
#include <dev/atkbdc/psm.h>

/*
 * Driver specific options: the following options may be set by
 * `options' statements in the kernel configuration file.
 */

/* debugging */
#ifndef PSM_DEBUG
#define PSM_DEBUG	0	/*
				 * logging: 0: none, 1: brief, 2: verbose
				 *          3: sync errors, 4: all packets
				 */
#endif
#define VLOG(level, args) 	\
do {				\
	if (verbose >= level)	\
		log args;	\
} while (0)

#ifndef PSM_INPUT_TIMEOUT
#define PSM_INPUT_TIMEOUT	2000000	/* 2 sec */
#endif

/* end of driver specific options */

#define PSMCPNP_DRIVER_NAME	"psmcpnp"

/* input queue */
#define PSM_BUFSIZE		960
#define PSM_SMALLBUFSIZE	240

/* some macros */
#define PSM_UNIT(dev)		(minor(dev) >> 1)
#define PSM_NBLOCKIO(dev)	(minor(dev) & 1)
#define PSM_MKMINOR(unit,block)	(((unit) << 1) | ((block) ? 0:1))

/* ring buffer */
typedef struct ringbuf {
    int           count;	/* # of valid elements in the buffer */
    int           head;		/* head pointer */
    int           tail;		/* tail poiner */
    unsigned char buf[PSM_BUFSIZE];
} ringbuf_t;

/* data buffer */
typedef struct packetbuf {
    unsigned char ipacket[16];	/* interim input buffer */
    int           inputbytes;	/* # of bytes in the input buffer */
} packetbuf_t;

#ifndef PSM_PACKETQUEUE
#define PSM_PACKETQUEUE	128
#endif

/* driver control block */
struct psm_softc {		/* Driver status information */
    int		  unit;
    struct selinfo rsel;	/* Process selecting for Input */
    unsigned char state;	/* Mouse driver state */
    int           config;	/* driver configuration flags */
    int           flags;	/* other flags */
    KBDC          kbdc;		/* handle to access the keyboard controller */
    struct resource *intr;	/* IRQ resource */
    void	  *ih;		/* interrupt handle */
    ringbuf_t     queue;	/* mouse status queue */
    packetbuf_t   pqueue[PSM_PACKETQUEUE];	/* mouse data queue */
    int           pqueue_start; /* start of data in queue */
    int           pqueue_end;   /* end of data in queue */
    int           button;	/* the latest button state */
    int		  syncerrors;	/* # of bytes discarded searching for sync */
    int		  pkterrors;	/* # of packets failed during quaranteen. */
    struct timeval inputtimeout;
    struct timeval lastsoftintr;	/* time of last soft interrupt */
    struct timeval lastinputerr;	/* time last sync error happened */
    int		  watchdog;		/* watchdog timer flag */
    struct callout_handle callout;	/* watchdog timer call out */
    struct callout_handle softcallout;	/* buffer timer call out */
    struct cdev *dev;
    struct cdev *bdev;
    int           lasterr;
    int           cmdcount;
    mousemode_t   mode;
};

static devclass_t psm_devclass;
#define PSM_SOFTC(unit)	((struct psm_softc*)devclass_get_softc(psm_devclass, unit))

/* driver state flags (state) */
#define PSM_VALID		0x80
#define PSM_OPEN		1	/* Device is open */
#define PSM_ASLP		2	/* Waiting for mouse data */
#define PSM_SOFTARMED		4	/* Software interrupt armed */
#define PSM_NEED_SYNCBITS	8	/* Set syncbits using next data pkt */

/* driver configuration flags (config) */
#define PSM_CONFIG_RESOLUTION	0x000f	/* resolution */
#define PSM_CONFIG_ACCEL	0x00f0  /* acceleration factor */
#define PSM_CONFIG_NOCHECKSYNC	0x0100  /* disable sync. test */
#define PSM_CONFIG_NOIDPROBE	0x0200  /* disable mouse model probe */
#define PSM_CONFIG_NORESET	0x0400  /* don't reset the mouse */
#define PSM_CONFIG_FORCETAP	0x0800  /* assume `tap' action exists */
#define PSM_CONFIG_IGNPORTERROR	0x1000  /* ignore error in aux port test */
#define PSM_CONFIG_HOOKRESUME	0x2000	/* hook the system resume event */
#define PSM_CONFIG_INITAFTERSUSPEND 0x4000 /* init the device at the resume event */
#define PSM_CONFIG_SYNCHACK	0x8000 /* enable `out-of-sync' hack */

#define PSM_CONFIG_FLAGS	(PSM_CONFIG_RESOLUTION 		\
				    | PSM_CONFIG_ACCEL		\
				    | PSM_CONFIG_NOCHECKSYNC	\
				    | PSM_CONFIG_SYNCHACK	\
				    | PSM_CONFIG_NOIDPROBE	\
				    | PSM_CONFIG_NORESET	\
				    | PSM_CONFIG_FORCETAP	\
				    | PSM_CONFIG_IGNPORTERROR	\
				    | PSM_CONFIG_HOOKRESUME	\
				    | PSM_CONFIG_INITAFTERSUSPEND)

/* other flags (flags) */
#define PSM_FLAGS_FINGERDOWN	0x0001 /* VersaPad finger down */

/* Tunables */
static int verbose = PSM_DEBUG;
TUNABLE_INT("debug.psm.loglevel", &verbose);

/* function prototypes */
static void psmidentify(driver_t *, device_t);
static int psmprobe(device_t);
static int psmattach(device_t);
static int psmdetach(device_t);
static int psmresume(device_t);

static d_open_t psmopen;
static d_close_t psmclose;
static d_read_t psmread;
static d_ioctl_t psmioctl;
static d_poll_t psmpoll;

static int enable_aux_dev(KBDC);
static int disable_aux_dev(KBDC);
static int get_mouse_status(KBDC, int *, int, int);
static void recover_from_error(KBDC);
static int restore_controller(KBDC, int);
static int doinitialize(struct psm_softc *);
static int doopen(struct psm_softc *, int);
static int reinitialize(struct psm_softc *, int);
static void psmsoftintr(void *);
static void psmintr(void *);
static void psmtimeout(void *);
static int timeelapsed(const struct timeval *,
    int, int, const struct timeval *);
static void dropqueue(struct psm_softc *);
static void flushpackets(struct psm_softc *);

/* device driver declarateion */
static device_method_t psm_methods[] = {
	/* Device interface */
	DEVMETHOD(device_identify,	psmidentify),
	DEVMETHOD(device_probe,		psmprobe),
	DEVMETHOD(device_attach,	psmattach),
	DEVMETHOD(device_detach,	psmdetach),
	DEVMETHOD(device_resume,	psmresume),

	{ 0, 0 }
};

static driver_t psm_driver = {
    PSM_DRIVER_NAME,
    psm_methods,
    sizeof(struct psm_softc),
};


static struct cdevsw psm_cdevsw = {
	.d_version =	D_VERSION,
	.d_flags =	D_NEEDGIANT,
	.d_open =	psmopen,
	.d_close =	psmclose,
	.d_read =	psmread,
	.d_ioctl =	psmioctl,
	.d_poll =	psmpoll,
	.d_name =	PSM_DRIVER_NAME,
};

/* device I/O routines */
static int
enable_aux_dev(KBDC kbdc)
{
    int res;

    res = send_aux_command(kbdc, PSMC_ENABLE_DEV);
    VLOG(2, (LOG_DEBUG, "psm: ENABLE_DEV return code:%04x\n", res));

    return (res == PSM_ACK);
}

static int
disable_aux_dev(KBDC kbdc)
{
    int res;

    res = send_aux_command(kbdc, PSMC_DISABLE_DEV);
    VLOG(2, (LOG_DEBUG, "psm: DISABLE_DEV return code:%04x\n", res));

    return (res == PSM_ACK);
}

static int
get_mouse_status(KBDC kbdc, int *status, int flag, int len)
{
    int cmd;
    int res;
    int i;

    switch (flag) {
    case 0:
    default:
	cmd = PSMC_SEND_DEV_STATUS;
	break;
    case 1:
	cmd = PSMC_SEND_DEV_DATA;
	break;
    }
    empty_aux_buffer(kbdc, 5);
    res = send_aux_command(kbdc, cmd);
    VLOG(2, (LOG_DEBUG, "psm: SEND_AUX_DEV_%s return code:%04x\n", 
	 (flag == 1) ? "DATA" : "STATUS", res));
    if (res != PSM_ACK)
        return 0;

    for (i = 0; i < len; ++i) {
        status[i] = read_aux_data(kbdc);
	if (status[i] < 0)
	    break;
    }

    VLOG(1, (LOG_DEBUG, "psm: %s %02x %02x %02x\n",
         (flag == 1) ? "data" : "status", status[0], status[1], status[2]));

    return i;
}

/* misc subroutines */

static void
recover_from_error(KBDC kbdc)
{
    /* discard anything left in the output buffer */
    empty_both_buffers(kbdc, 10);

    /*
     * NOTE: somehow diagnostic and keyboard port test commands bring the
     * keyboard back.
     */
    if (!test_controller(kbdc)) 
        log(LOG_ERR, "psm: keyboard controller failed.\n");
    /* if there isn't a keyboard in the system, the following error is OK */
    if (test_kbd_port(kbdc) != 0)
	VLOG(1, (LOG_ERR, "psm: keyboard port failed.\n"));
}

static int
restore_controller(KBDC kbdc, int command_byte)
{
    empty_both_buffers(kbdc, 10);

    if (!set_controller_command_byte(kbdc, 0xff, command_byte)) {
	log(LOG_ERR, "psm: failed to restore the keyboard controller "
		     "command byte.\n");
	empty_both_buffers(kbdc, 10);
	return FALSE;
    } else {
	empty_both_buffers(kbdc, 10);
	return TRUE;
    }
}

/* 
 * Re-initialize the aux port and device. The aux port must be enabled
 * and its interrupt must be disabled before calling this routine. 
 * The aux device will be disabled before returning.
 * The keyboard controller must be locked via `kbdc_lock()' before
 * calling this routine.
 */
static int
doinitialize(struct psm_softc *sc)
{
    KBDC kbdc = sc->kbdc;
    int stat[3];
    int i;

    switch((i = test_aux_port(kbdc))) {
    case 1:	/* ignore these errors */
    case 2:
    case 3:
    case PSM_ACK:
	if (verbose)
	    log(LOG_DEBUG, "psm%d: strange result for test aux port (%d).\n",
	        sc->unit, i);
	/* FALLTHROUGH */
    case 0:	/* no error */
    	break;
    case -1: 	/* time out */
    default: 	/* error */
    	recover_from_error(kbdc);
	if (sc->config & PSM_CONFIG_IGNPORTERROR)
	    break;
    	log(LOG_ERR, "psm%d: the aux port is not functioning (%d).\n",
    	    sc->unit, i);
    	return FALSE;
    }

    if (sc->config & PSM_CONFIG_NORESET) {
	/* 
	 * Don't try to reset the pointing device.  It may possibly be
	 * left in the unknown state, though...
	 */
    } else {
	/* 
	 * NOTE: some controllers appears to hang the `keyboard' when
	 * the aux port doesn't exist and `PSMC_RESET_DEV' is issued. 
	 */
	if (!reset_aux_dev(kbdc)) {
            recover_from_error(kbdc);
            log(LOG_ERR, "psm%d: failed to reset the aux device.\n", sc->unit);
            return FALSE;
	}
    }

    /* 
     * both the aux port and the aux device is functioning, see
     * if the device can be enabled. 
     */
    if (!enable_aux_dev(kbdc) || !disable_aux_dev(kbdc)) {
        log(LOG_ERR, "psm%d: failed to enable the aux device.\n", sc->unit);
        return FALSE;
    }
    empty_both_buffers(kbdc, 10);	/* remove stray data if any */

    /* Record sync on the next data packet we see. */
    sc->flags |= PSM_NEED_SYNCBITS;

    /* just check the status of the mouse */
    if (get_mouse_status(kbdc, stat, 0, 3) < 3)
        log(LOG_DEBUG, "psm%d: failed to get status (doinitialize).\n",
	    sc->unit);

    return TRUE;
}

static int
doopen(struct psm_softc *sc, int command_byte)
{
    int stat[3];

    /* enable the mouse device */
    if (!enable_aux_dev(sc->kbdc)) {
	/* MOUSE ERROR: failed to enable the mouse because:
	 * 1) the mouse is faulty,
	 * 2) the mouse has been removed(!?)
	 * In the latter case, the keyboard may have hung, and need 
	 * recovery procedure...
	 */
	recover_from_error(sc->kbdc);
        {
            restore_controller(sc->kbdc, command_byte);
	    /* mark this device is no longer available */
	    sc->state &= ~PSM_VALID;	
	    log(LOG_ERR, "psm%d: failed to enable the device (doopen).\n",
		sc->unit);
	    return (EIO);
	}
    }

    if (get_mouse_status(sc->kbdc, stat, 0, 3) < 3) 
        log(LOG_DEBUG, "psm%d: failed to get status (doopen).\n", sc->unit);

    /* enable the aux port and interrupt */
    if (!set_controller_command_byte(sc->kbdc, 
	    kbdc_get_device_mask(sc->kbdc),
	    (command_byte & KBD_KBD_CONTROL_BITS)
		| KBD_ENABLE_AUX_PORT | KBD_ENABLE_AUX_INT)) {
	/* CONTROLLER ERROR */
	disable_aux_dev(sc->kbdc);
        restore_controller(sc->kbdc, command_byte);
	log(LOG_ERR, "psm%d: failed to enable the aux interrupt (doopen).\n",
	    sc->unit);
	return (EIO);
    }

    /* start the watchdog timer */
    sc->watchdog = FALSE;
    sc->callout = timeout(psmtimeout, (void *)(uintptr_t)sc, hz*2);

    return (0);
}

static int
reinitialize(struct psm_softc *sc, int doinit)
{
    int err;
    int c;
    int s;

    /* don't let anybody mess with the aux device */
    if (!kbdc_lock(sc->kbdc, TRUE))
	return (EIO);
    s = spltty();

    /* block our watchdog timer */
    sc->watchdog = FALSE;
    untimeout(psmtimeout, (void *)(uintptr_t)sc, sc->callout);
    callout_handle_init(&sc->callout);

    /* save the current controller command byte */
    empty_both_buffers(sc->kbdc, 10);
    c = get_controller_command_byte(sc->kbdc);
    VLOG(2, (LOG_DEBUG, "psm%d: current command byte: %04x (reinitialize).\n", 
	 sc->unit, c));

    /* enable the aux port but disable the aux interrupt and the keyboard */
    if ((c == -1) || !set_controller_command_byte(sc->kbdc,
	    kbdc_get_device_mask(sc->kbdc),
  	    KBD_DISABLE_KBD_PORT | KBD_DISABLE_KBD_INT
	        | KBD_ENABLE_AUX_PORT | KBD_DISABLE_AUX_INT)) {
        /* CONTROLLER ERROR */
	splx(s);
        kbdc_lock(sc->kbdc, FALSE);
	log(LOG_ERR, "psm%d: unable to set the command byte (reinitialize).\n",
	    sc->unit);
	return (EIO);
    }

    /* flush any data */
    if (sc->state & PSM_VALID) {
	disable_aux_dev(sc->kbdc);	/* this may fail; but never mind... */
	empty_aux_buffer(sc->kbdc, 10);
    }
    flushpackets(sc);
    sc->syncerrors = 0;
    sc->pkterrors = 0;
    memset(&sc->lastinputerr, 0, sizeof(sc->lastinputerr));

    /* try to detect the aux device; are you still there? */
    err = 0;
    if (doinit) {
	if (doinitialize(sc)) {
	    /* yes */
	    sc->state |= PSM_VALID;
	} else {
	    /* the device has gone! */
	    restore_controller(sc->kbdc, c);
	    sc->state &= ~PSM_VALID;
	    log(LOG_ERR, "psm%d: the aux device has gone! (reinitialize).\n",
		sc->unit);
	    err = ENXIO;
	}
    }
    splx(s);

    /* restore the driver state */
    if ((sc->state & PSM_OPEN) && (err == 0)) {
        /* enable the aux device and the port again */
	err = doopen(sc, c);
	if (err != 0) 
	    log(LOG_ERR, "psm%d: failed to enable the device (reinitialize).\n",
		sc->unit);
    } else {
        /* restore the keyboard port and disable the aux port */
        if (!set_controller_command_byte(sc->kbdc, 
                kbdc_get_device_mask(sc->kbdc),
                (c & KBD_KBD_CONTROL_BITS)
                    | KBD_DISABLE_AUX_PORT | KBD_DISABLE_AUX_INT)) {
            /* CONTROLLER ERROR */
            log(LOG_ERR,
                "psm%d: failed to disable the aux port (reinitialize).\n",
                sc->unit);
            err = EIO;
	}
    }

    kbdc_lock(sc->kbdc, FALSE);
    return (err);
}

/* psm driver entry points */

static void
psmidentify(driver_t *driver, device_t parent)
{
    device_t psmc;
    device_t psm;
    u_long irq;
    int unit;

    unit = device_get_unit(parent);

    /* always add at least one child */
    psm = BUS_ADD_CHILD(parent, KBDC_RID_AUX, driver->name, unit);
    if (psm == NULL)
	return;

    irq = bus_get_resource_start(psm, SYS_RES_IRQ, KBDC_RID_AUX);
    if (irq > 0)
	return;

    /*
     * If the PS/2 mouse device has already been reported by ACPI or
     * PnP BIOS, obtain the IRQ resource from it.
     * (See psmcpnp_attach() below.)
     */
    psmc = device_find_child(device_get_parent(parent),
			     PSMCPNP_DRIVER_NAME, unit);
    if (psmc == NULL)
	return;
    irq = bus_get_resource_start(psmc, SYS_RES_IRQ, 0);
    if (irq <= 0)
	return;
    bus_set_resource(psm, SYS_RES_IRQ, KBDC_RID_AUX, irq, 1);
}

#define endprobe(v)	do {   if (bootverbose)				\
				--verbose;   				\
                            kbdc_set_device_mask(sc->kbdc, mask);	\
			    kbdc_lock(sc->kbdc, FALSE);			\
			    return (v);	     				\
			} while (0)

static int
psmprobe(device_t dev)
{
    int unit = device_get_unit(dev);
    struct psm_softc *sc = device_get_softc(dev);
    int stat[3];
    int command_byte;
    int mask;
    int rid;
    int i;

#if 0
    kbdc_debug(TRUE);
#endif

    /* see if IRQ is available */
    rid = KBDC_RID_AUX;
    sc->intr = bus_alloc_resource_any(dev, SYS_RES_IRQ, &rid,
				      RF_SHAREABLE | RF_ACTIVE);
    if (sc->intr == NULL) {
	if (bootverbose)
            device_printf(dev, "unable to allocate IRQ\n");
        return (ENXIO);
    }
    bus_release_resource(dev, SYS_RES_IRQ, rid, sc->intr);

    sc->unit = unit;
    sc->kbdc = atkbdc_open(device_get_unit(device_get_parent(dev)));
    sc->config = device_get_flags(dev) & PSM_CONFIG_FLAGS;
    /* XXX: for backward compatibility */
#if defined(PSM_HOOKRESUME) || defined(PSM_HOOKAPM)
    sc->config |= 
#ifdef PSM_RESETAFTERSUSPEND
	PSM_CONFIG_HOOKRESUME | PSM_CONFIG_INITAFTERSUSPEND;
#else
	PSM_CONFIG_HOOKRESUME;
#endif
#endif /* PSM_HOOKRESUME | PSM_HOOKAPM */
    sc->flags = 0;
    if (bootverbose)
        ++verbose;

    device_set_desc(dev, "(new) PS/2 Mouse");

    if (!kbdc_lock(sc->kbdc, TRUE)) {
        printf("psm%d: unable to lock the controller.\n", unit);
        if (bootverbose)
            --verbose;
	return (ENXIO);
    }

    /*
     * NOTE: two bits in the command byte controls the operation of the
     * aux port (mouse port): the aux port disable bit (bit 5) and the aux
     * port interrupt (IRQ 12) enable bit (bit 2).
     */

    /* discard anything left after the keyboard initialization */
    empty_both_buffers(sc->kbdc, 10);

    /* save the current command byte; it will be used later */
    mask = kbdc_get_device_mask(sc->kbdc) & ~KBD_AUX_CONTROL_BITS;
    command_byte = get_controller_command_byte(sc->kbdc);
    if (verbose) 
        printf("psm%d: current command byte:%04x\n", unit, command_byte);
    if (command_byte == -1) {
        /* CONTROLLER ERROR */
        printf("psm%d: unable to get the current command byte value.\n",
            unit);
        endprobe(ENXIO);
    }

    /*
     * disable the keyboard port while probing the aux port, which must be
     * enabled during this routine
     */
    if (!set_controller_command_byte(sc->kbdc,
	    KBD_KBD_CONTROL_BITS | KBD_AUX_CONTROL_BITS,
  	    KBD_DISABLE_KBD_PORT | KBD_DISABLE_KBD_INT
                | KBD_ENABLE_AUX_PORT | KBD_DISABLE_AUX_INT)) {
        /* 
	 * this is CONTROLLER ERROR; I don't know how to recover 
         * from this error... 
	 */
        restore_controller(sc->kbdc, command_byte);
        printf("psm%d: unable to set the command byte.\n", unit);
        endprobe(ENXIO);
    }
    write_controller_command(sc->kbdc, KBDC_ENABLE_AUX_PORT);

    /*
     * NOTE: `test_aux_port()' is designed to return with zero if the aux
     * port exists and is functioning. However, some controllers appears
     * to respond with zero even when the aux port doesn't exist. (It may
     * be that this is only the case when the controller DOES have the aux
     * port but the port is not wired on the motherboard.) The keyboard
     * controllers without the port, such as the original AT, are
     * supporsed to return with an error code or simply time out. In any
     * case, we have to continue probing the port even when the controller
     * passes this test.
     *
     * XXX: some controllers erroneously return the error code 1, 2 or 3
     * when it has the perfectly functional aux port. We have to ignore
     * this error code. Even if the controller HAS error with the aux
     * port, it will be detected later...
     * XXX: another incompatible controller returns PSM_ACK (0xfa)...
     */
    switch ((i = test_aux_port(sc->kbdc))) {
    case 1:	   /* ignore these errors */
    case 2:
    case 3:
    case PSM_ACK:
        if (verbose)
	    printf("psm%d: strange result for test aux port (%d).\n",
	        unit, i);
	/* FALLTHROUGH */
    case 0:        /* no error */
        break;
    case -1:        /* time out */
    default:        /* error */
        recover_from_error(sc->kbdc);
	if (sc->config & PSM_CONFIG_IGNPORTERROR)
	    break;
        restore_controller(sc->kbdc, command_byte);
        if (verbose)
            printf("psm%d: the aux port is not functioning (%d).\n",
                unit, i);
        endprobe(ENXIO);
    }

    if (sc->config & PSM_CONFIG_NORESET) {
	/* 
	 * Don't try to reset the pointing device.  It may possibly be
	 * left in the unknown state, though...
	 */
    } else {
	/*
	 * NOTE: some controllers appears to hang the `keyboard' when the aux
	 * port doesn't exist and `PSMC_RESET_DEV' is issued.
	 *
	 * Attempt to reset the controller twice -- this helps
	 * pierce through some KVM switches. The second reset
	 * is non-fatal.
	 */
	if (!reset_aux_dev(sc->kbdc)) {
            recover_from_error(sc->kbdc);
            restore_controller(sc->kbdc, command_byte);
            if (verbose)
        	printf("psm%d: failed to reset the aux device.\n", unit);
            endprobe(ENXIO);
	} else if (!reset_aux_dev(sc->kbdc)) {
	    recover_from_error(sc->kbdc);
	    if (verbose >= 2)
        	printf("psm%d: failed to reset the aux device (2).\n",
        	    unit);
	}
    }

    /*
     * both the aux port and the aux device is functioning, see if the
     * device can be enabled. NOTE: when enabled, the device will start
     * sending data; we shall immediately disable the device once we know
     * the device can be enabled.
     */
    if (!enable_aux_dev(sc->kbdc) || !disable_aux_dev(sc->kbdc)) {
        /* MOUSE ERROR */
	recover_from_error(sc->kbdc);
	restore_controller(sc->kbdc, command_byte);
	if (verbose)
	    printf("psm%d: failed to enable the aux device.\n", unit);
        endprobe(ENXIO);
    }

    /* Record sync on the next data packet we see. */
    sc->flags |= PSM_NEED_SYNCBITS;

    /* just check the status of the mouse */
    /* 
     * NOTE: XXX there are some arcane controller/mouse combinations out 
     * there, which hung the controller unless there is data transmission 
     * after ACK from the mouse.
     */
    if (get_mouse_status(sc->kbdc, stat, 0, 3) < 3) {
        printf("psm%d: failed to get status.\n", unit);
    }

    /* disable the aux port for now... */
    if (!set_controller_command_byte(sc->kbdc, 
	    KBD_KBD_CONTROL_BITS | KBD_AUX_CONTROL_BITS,
            (command_byte & KBD_KBD_CONTROL_BITS)
                | KBD_DISABLE_AUX_PORT | KBD_DISABLE_AUX_INT)) {
        /* 
	 * this is CONTROLLER ERROR; I don't know the proper way to 
         * recover from this error... 
	 */
        restore_controller(sc->kbdc, command_byte);
        printf("psm%d: unable to set the command byte.\n", unit);
        endprobe(ENXIO);
    }

    /* done */
    kbdc_set_device_mask(sc->kbdc, mask | KBD_AUX_CONTROL_BITS);
    kbdc_lock(sc->kbdc, FALSE);
    return (0);
}

static int
psmattach(device_t dev)
{
    int unit = device_get_unit(dev);
    struct psm_softc *sc = device_get_softc(dev);
    int error;
    int rid;

    /* Setup initial state */
    sc->state = PSM_VALID;
    callout_handle_init(&sc->callout);

    /* Setup our interrupt handler */
    rid = KBDC_RID_AUX;
    sc->intr = bus_alloc_resource_any(dev, SYS_RES_IRQ, &rid,
				      RF_SHAREABLE | RF_ACTIVE);
    if (sc->intr == NULL)
	return (ENXIO);
    error = bus_setup_intr(dev, sc->intr, INTR_TYPE_TTY, psmintr, sc, &sc->ih);
    if (error) {
	bus_release_resource(dev, SYS_RES_IRQ, rid, sc->intr);
	return (error);
    }

    /* Done */
    sc->dev = make_dev(&psm_cdevsw, PSM_MKMINOR(unit, FALSE), 0, 0, 0666,
		       "psm%d", unit);
    sc->bdev = make_dev(&psm_cdevsw, PSM_MKMINOR(unit, TRUE), 0, 0, 0666,
			"bpsm%d", unit);

    printf("psm%d: attached\n", unit);

    if (bootverbose)
        --verbose;

    return (0);
}

static int
psmdetach(device_t dev)
{
    struct psm_softc *sc;
    int rid;

    sc = device_get_softc(dev);
    if (sc->state & PSM_OPEN)
	return EBUSY;

    rid = KBDC_RID_AUX;
    bus_teardown_intr(dev, sc->intr, sc->ih);
    bus_release_resource(dev, SYS_RES_IRQ, rid, sc->intr);

    destroy_dev(sc->dev);
    destroy_dev(sc->bdev);

    return 0;
}

static int
psmopen(struct cdev *dev, int flag, int fmt, struct thread *td)
{
    int unit = PSM_UNIT(dev);
    struct psm_softc *sc;
    int command_byte;
    int err;
    int s;

    /* Get device data */
    sc = PSM_SOFTC(unit);
    if ((sc == NULL) || (sc->state & PSM_VALID) == 0)
	/* the device is no longer valid/functioning */
        return (ENXIO);

    /* Disallow multiple opens */
    if (sc->state & PSM_OPEN)
        return (EBUSY);

    device_busy(devclass_get_device(psm_devclass, unit));

    /* Initialize state */
    sc->watchdog = FALSE;

    /* flush the event queue */
    sc->queue.count = 0;
    sc->queue.head = 0;
    sc->queue.tail = 0;
    sc->button = 0;
    sc->pqueue_start = 0;
    sc->pqueue_end = 0;

    /* empty input buffer */
    flushpackets(sc);
    sc->syncerrors = 0;
    sc->pkterrors = 0;

    /* don't let timeout routines in the keyboard driver to poll the kbdc */
    if (!kbdc_lock(sc->kbdc, TRUE))
	return (EIO);

    /* save the current controller command byte */
    s = spltty();
    command_byte = get_controller_command_byte(sc->kbdc);

    /* enable the aux port and temporalily disable the keyboard */
    if ((command_byte == -1) 
        || !set_controller_command_byte(sc->kbdc,
	    kbdc_get_device_mask(sc->kbdc),
  	    KBD_DISABLE_KBD_PORT | KBD_DISABLE_KBD_INT
	        | KBD_ENABLE_AUX_PORT | KBD_DISABLE_AUX_INT)) {
        /* CONTROLLER ERROR; do you know how to get out of this? */
        kbdc_lock(sc->kbdc, FALSE);
	splx(s);
	log(LOG_ERR, "psm%d: unable to set the command byte (psmopen).\n",
	    unit);
	return (EIO);
    }
    /* 
     * Now that the keyboard controller is told not to generate 
     * the keyboard and mouse interrupts, call `splx()' to allow 
     * the other tty interrupts. The clock interrupt may also occur, 
     * but timeout routines will be blocked by the poll flag set 
     * via `kbdc_lock()'
     */
    splx(s);
  
    /* enable the mouse device */
    err = doopen(sc, command_byte);

    /* done */
    if (err == 0) 
        sc->state |= PSM_OPEN;
    kbdc_lock(sc->kbdc, FALSE);
    return (err);
}

static int
psmclose(struct cdev *dev, int flag, int fmt, struct thread *td)
{
    int unit = PSM_UNIT(dev);
    struct psm_softc *sc = PSM_SOFTC(unit);
    int stat[3];
    int command_byte;
    int s;

    /* don't let timeout routines in the keyboard driver to poll the kbdc */
    if (!kbdc_lock(sc->kbdc, TRUE))
	return (EIO);

    /* save the current controller command byte */
    s = spltty();
    command_byte = get_controller_command_byte(sc->kbdc);
    if (command_byte == -1) {
        kbdc_lock(sc->kbdc, FALSE);
	splx(s);
	return (EIO);
    }

    /* disable the aux interrupt and temporalily disable the keyboard */
    if (!set_controller_command_byte(sc->kbdc, 
	    kbdc_get_device_mask(sc->kbdc),
  	    KBD_DISABLE_KBD_PORT | KBD_DISABLE_KBD_INT
	        | KBD_ENABLE_AUX_PORT | KBD_DISABLE_AUX_INT)) {
	log(LOG_ERR, "psm%d: failed to disable the aux int (psmclose).\n",
	    unit);
	/* CONTROLLER ERROR;
	 * NOTE: we shall force our way through. Because the only
	 * ill effect we shall see is that we may not be able
	 * to read ACK from the mouse, and it doesn't matter much 
	 * so long as the mouse will accept the DISABLE command.
	 */
    }
    splx(s);

    /* stop the watchdog timer */
    untimeout(psmtimeout, (void *)(uintptr_t)sc, sc->callout);
    callout_handle_init(&sc->callout);

    /* remove anything left in the output buffer */
    empty_aux_buffer(sc->kbdc, 10);

    /* disable the aux device, port and interrupt */
    if (sc->state & PSM_VALID) {
        if (!disable_aux_dev(sc->kbdc)) {
	    /* MOUSE ERROR; 
	     * NOTE: we don't return error and continue, pretending 
	     * we have successfully disabled the device. It's OK because 
	     * the interrupt routine will discard any data from the mouse
	     * hereafter. 
	     */
	    log(LOG_ERR, "psm%d: failed to disable the device (psmclose).\n",
		unit);
        }

        if (get_mouse_status(sc->kbdc, stat, 0, 3) < 3)
            log(LOG_DEBUG, "psm%d: failed to get status (psmclose).\n", unit);
    }

    if (!set_controller_command_byte(sc->kbdc, 
	    kbdc_get_device_mask(sc->kbdc),
	    (command_byte & KBD_KBD_CONTROL_BITS)
	        | KBD_DISABLE_AUX_PORT | KBD_DISABLE_AUX_INT)) {
	/* CONTROLLER ERROR; 
	 * we shall ignore this error; see the above comment.
	 */
	log(LOG_ERR, "psm%d: failed to disable the aux port (psmclose).\n",
	    unit);
    }

    /* remove anything left in the output buffer */
    empty_aux_buffer(sc->kbdc, 10);

    /* close is almost always successful */
    sc->state &= ~PSM_OPEN;
    kbdc_lock(sc->kbdc, FALSE);
    device_unbusy(devclass_get_device(psm_devclass, unit));
    return (0);
}

static int
psmread(struct cdev *dev, struct uio *uio, int flag)
{
    register struct psm_softc *sc = PSM_SOFTC(PSM_UNIT(dev));
    unsigned char buf[PSM_SMALLBUFSIZE];
    int error = 0;
    int s;
    int l;

    if ((sc->state & PSM_VALID) == 0)
	return EIO;

    /* block until mouse activity occured */
    s = spltty();
    while (sc->queue.count <= 0) {
        if (PSM_NBLOCKIO(dev)) {
            splx(s);
            return EWOULDBLOCK;
        }
        sc->state |= PSM_ASLP;
        error = tsleep( sc, PZERO | PCATCH, "psmrea", 0);
        sc->state &= ~PSM_ASLP;
        if (error) {
            splx(s);
            return error;
        } else if ((sc->state & PSM_VALID) == 0) {
            /* the device disappeared! */
            splx(s);
            return EIO;
	}
    }
    splx(s);

    /* copy data to the user land */
    while ((sc->queue.count > 0) && (uio->uio_resid > 0)) {
        s = spltty();
	l = imin(sc->queue.count, uio->uio_resid);
	if (l > sizeof(buf))
	    l = sizeof(buf);
	if (l > sizeof(sc->queue.buf) - sc->queue.head) {
	    bcopy(&sc->queue.buf[sc->queue.head], &buf[0], 
		sizeof(sc->queue.buf) - sc->queue.head);
	    bcopy(&sc->queue.buf[0], 
		&buf[sizeof(sc->queue.buf) - sc->queue.head],
		l - (sizeof(sc->queue.buf) - sc->queue.head));
	} else {
	    bcopy(&sc->queue.buf[sc->queue.head], &buf[0], l);
	}
	sc->queue.count -= l;
	sc->queue.head = (sc->queue.head + l) % sizeof(sc->queue.buf);
        splx(s);
        error = uiomove(buf, l, uio);
        if (error)
	    break;
    }

    return error;
}

static int
block_mouse_data(struct psm_softc *sc, int *c)
{
    int s;

    if (!kbdc_lock(sc->kbdc, TRUE)) 
	return EIO;

    s = spltty();
    *c = get_controller_command_byte(sc->kbdc);
    if ((*c == -1) 
	|| !set_controller_command_byte(sc->kbdc, 
	    kbdc_get_device_mask(sc->kbdc),
            KBD_DISABLE_KBD_PORT | KBD_DISABLE_KBD_INT
                | KBD_ENABLE_AUX_PORT | KBD_DISABLE_AUX_INT)) {
        /* this is CONTROLLER ERROR */
	splx(s);
        kbdc_lock(sc->kbdc, FALSE);
	return EIO;
    }

    /* 
     * The device may be in the middle of status data transmission.
     * The transmission will be interrupted, thus, incomplete status 
     * data must be discarded. Although the aux interrupt is disabled 
     * at the keyboard controller level, at most one aux interrupt 
     * may have already been pending and a data byte is in the 
     * output buffer; throw it away. Note that the second argument 
     * to `empty_aux_buffer()' is zero, so that the call will just 
     * flush the internal queue.
     * `psmintr()' will be invoked after `splx()' if an interrupt is
     * pending; it will see no data and returns immediately.
     */
    empty_aux_buffer(sc->kbdc, 0);	/* flush the queue */
    read_aux_data_no_wait(sc->kbdc);	/* throw away data if any */
    flushpackets(sc);
    splx(s);

    return 0;
}

static void
dropqueue(struct psm_softc *sc)
{

    	sc->queue.count = 0;
   	sc->queue.head = 0;
    	sc->queue.tail = 0;
	if ((sc->state & PSM_SOFTARMED) != 0) {
		sc->state &= ~PSM_SOFTARMED;
		untimeout(psmsoftintr, (void *)(uintptr_t)sc, sc->softcallout);
	}
	sc->pqueue_start = sc->pqueue_end;
}

static void
flushpackets(struct psm_softc *sc)
{

	dropqueue(sc);
	bzero(&sc->pqueue, sizeof(sc->pqueue));
}

static int
unblock_mouse_data(struct psm_softc *sc, int c)
{
    int error = 0;

    /* 
     * We may have seen a part of status data during `set_mouse_XXX()'.
     * they have been queued; flush it.
     */
    empty_aux_buffer(sc->kbdc, 0);

    /* restore ports and interrupt */
    if (!set_controller_command_byte(sc->kbdc, 
            kbdc_get_device_mask(sc->kbdc),
	    c & (KBD_KBD_CONTROL_BITS | KBD_AUX_CONTROL_BITS))) {
        /* CONTROLLER ERROR; this is serious, we may have
         * been left with the inaccessible keyboard and
         * the disabled mouse interrupt. 
         */
        error = EIO;
    }

    kbdc_lock(sc->kbdc, FALSE);
    return error;
}

static int
psmioctl(struct cdev *dev, u_long cmd, caddr_t addr, int flag, struct thread *td)
{
    struct psm_softc *sc = PSM_SOFTC(PSM_UNIT(dev));
    int command_byte;
    int error = 0;
    int s;

    /* Perform IOCTL command */
    switch (cmd) {
    case MOUSE_COMMAND:
	s = spltty();
	error = block_mouse_data(sc, &command_byte);
	if (error) {
	    uprintf("Error: %d\n", error);
	    return error;
	}

	if (((mousecommand_t *)addr)->data >= 0) {
	    int res;
	    res = send_aux_command_and_data(sc->kbdc,
					    ((mousecommand_t *)addr)->command,
					    ((mousecommand_t *)addr)->data);
	} else {
	    int res;
	    res = send_aux_command(sc->kbdc, ((mousecommand_t *)addr)->command);
	}

	DELAY(5000);

	{ /* XXX: This should be more cleanly written. Move 'x' decl up top? */
	    int x = 0;
	    for (x = 0; x < ((mousecommand_t *)addr)->responselen; x++) {
		((mousecommand_t *)addr)->response[x] = (char)read_aux_data(sc->kbdc);
	    }
	}

	unblock_mouse_data(sc, command_byte);
	splx(s);
	break;

    case MOUSE_SET_PACKETSIZE:
	s = spltty();
	sc->mode.packetsize = *(int *)addr;
	splx(s);
	break;

    case MOUSE_GET_PACKETSIZE:
	s = spltty();
	*(int *)addr = sc->mode.packetsize;
	splx(s);
	break;

    case MOUSE_SET_SYNCMASK:
	s = spltty();
	/* XXX: syncmask is an array becuase ums(4) wants it that way, for now */
	sc->mode.syncmask[0] = *(unsigned char *)addr;
	sc->flags |= PSM_NEED_SYNCBITS;
	uprintf("New syncmask: %02x\n", sc->mode.syncmask[0]);
	splx(s);
	break;

    case MOUSE_GET_SYNCMASK:
	s = spltty();
	/* XXX: ums likes syncmask to be an array */
	*(unsigned char *)addr = sc->mode.syncmask[0];
	splx(s);
	break;

    case MOUSE_SET_SYNCBITS:
	s = spltty();
	sc->mode.syncbits = *(unsigned char *)addr;
	splx(s);
	break;

    case MOUSE_GET_SYNCBITS:
	s = spltty();
	*(unsigned char *)addr = sc->mode.syncbits;
	splx(s);
	break;

    default:
	return ENOTTY;
    }

    return error;
}

static void
psmtimeout(void *arg)
{
    struct psm_softc *sc;
    int s;

    sc = (struct psm_softc *)arg;
    s = spltty();
    if (sc->watchdog && kbdc_lock(sc->kbdc, TRUE)) {
	VLOG(4, (LOG_DEBUG, "psm%d: lost interrupt?\n", sc->unit));
	psmintr(sc);
	kbdc_lock(sc->kbdc, FALSE);
    }
    sc->watchdog = TRUE;
    splx(s);
    sc->callout = timeout(psmtimeout, (void *)(uintptr_t)sc, hz);
}

/* Add all sysctls under the debug.psm and hw.psm nodes */
SYSCTL_NODE(_debug, OID_AUTO, psm, CTLFLAG_RD, 0, "ps/2 mouse");
SYSCTL_NODE(_hw, OID_AUTO, psm, CTLFLAG_RD, 0, "ps/2 mouse");

SYSCTL_INT(_debug_psm, OID_AUTO, loglevel, CTLFLAG_RW, &verbose, 0, "");

static int psmhz = 20;
SYSCTL_INT(_debug_psm, OID_AUTO, hz, CTLFLAG_RW, &psmhz, 0, "");
static int psmerrsecs = 2;
SYSCTL_INT(_debug_psm, OID_AUTO, errsecs, CTLFLAG_RW, &psmerrsecs, 0, "");
static int psmerrusecs = 0;
SYSCTL_INT(_debug_psm, OID_AUTO, errusecs, CTLFLAG_RW, &psmerrusecs, 0, "");
static int psmsecs = 0;
SYSCTL_INT(_debug_psm, OID_AUTO, secs, CTLFLAG_RW, &psmsecs, 0, "");
static int psmusecs = 500000;
SYSCTL_INT(_debug_psm, OID_AUTO, usecs, CTLFLAG_RW, &psmusecs, 0, "");
static int pkterrthresh = 2;
SYSCTL_INT(_debug_psm, OID_AUTO, pkterrthresh, CTLFLAG_RW, &pkterrthresh, 0, "");

//static int tap_threshold = PSM_TAP_THRESHOLD;
//SYSCTL_INT(_hw_psm, OID_AUTO, tap_threshold, CTLFLAG_RW, &tap_threshold, 0, "");
//static int tap_timeout = PSM_TAP_TIMEOUT;
//SYSCTL_INT(_hw_psm, OID_AUTO, tap_timeout, CTLFLAG_RW, &tap_timeout, 0, "");

static void
psmintr(void *arg)
{
    struct psm_softc *sc = arg;
    struct timeval now;
    int c;
    packetbuf_t *pb;

    /* read until there is nothing to read */
    while((c = read_aux_data_no_wait(sc->kbdc)) != -1) {
    
	//printf("Got data: %02x\n", c);
        pb = &sc->pqueue[sc->pqueue_end];
        /* discard the byte if the device is not open */
        if ((sc->state & PSM_OPEN) == 0)
            continue;
    
	getmicrouptime(&now);
	if ((pb->inputbytes > 0) && timevalcmp(&now, &sc->inputtimeout, >)) {
	    VLOG(3, (LOG_DEBUG, "psmintr: delay too long; "
		 "resetting byte count\n"));
	    pb->inputbytes = 0;
	    sc->syncerrors = 0;
	    sc->pkterrors = 0;
	}
	sc->inputtimeout.tv_sec = PSM_INPUT_TIMEOUT/1000000;
	sc->inputtimeout.tv_usec = PSM_INPUT_TIMEOUT%1000000;
	timevaladd(&sc->inputtimeout, &now);

        pb->ipacket[pb->inputbytes++] = c;
        if (pb->inputbytes < sc->mode.packetsize)
	    continue;

	VLOG(4, (LOG_DEBUG, "psmintr: %02x %02x %02x %02x %02x %02x\n",
	     pb->ipacket[0], pb->ipacket[1], pb->ipacket[2],
	     pb->ipacket[3], pb->ipacket[4], pb->ipacket[5]));

	c = pb->ipacket[0];

	if ((sc->flags & PSM_NEED_SYNCBITS) != 0) {
	    /* XXX: syncmask is an array becuase ums(4) wants it that way, for now */
	    sc->mode.syncbits = (c & sc->mode.syncmask[0]);
	    sc->flags &= ~PSM_NEED_SYNCBITS;
	    /* XXX: syncmask is an array becuase ums(4) wants it that way, for now */
	    VLOG(2, (LOG_DEBUG, "psmintr: Sync bytes now %04x,%04x [%02x]\n",
		     sc->mode.syncmask[0], sc->mode.syncbits, c));

	    /* XXX: syncmask is an array becuase ums(4) wants it that way, for now */
	} else if ((c & sc->mode.syncmask[0]) != sc->mode.syncbits) {
	    /* XXX: syncmask is an array becuase ums(4) wants it that way, for now */
	    VLOG(3, (LOG_DEBUG, "psmintr: out of sync (%04x != %04x) %d"
		     " cmds since last error.\n",
		     c & sc->mode.syncmask[0], sc->mode.syncbits,
		     sc->cmdcount - sc->lasterr));
	    sc->lasterr = sc->cmdcount;
	    /*
	     * The sync byte test is a weak measure of packet
	     * validity.  Conservatively discard any input yet
	     * to be seen by userland when we detect a sync
	     * error since there is a good chance some of
	     * the queued packets have undetected errors.
	     */
	    dropqueue(sc);
	    if (sc->syncerrors == 0)
		sc->pkterrors++;
	    ++sc->syncerrors;
	    sc->lastinputerr = now;
	    if (sc->syncerrors >= sc->mode.packetsize * 2 ||
		sc->pkterrors >= pkterrthresh) {

		/*
		 * If we've failed to find a single sync byte in 2
		 * packets worth of data, or we've seen persistent
		 * packet errors during the validation period,
		 * reinitialize the mouse in hopes of returning it
		 * to the expected mode.
		 */
		VLOG(3, (LOG_DEBUG, "psmintr: reset the mouse.\n"));
		reinitialize(sc, TRUE);
	    } else if (sc->syncerrors == sc->mode.packetsize) {

		/*
		 * Try a soft reset after searching for a sync
		 * byte through a packet length of bytes.
		 */
		VLOG(3, (LOG_DEBUG, "psmintr: re-enable the mouse.\n"));
		pb->inputbytes = 0;
		disable_aux_dev(sc->kbdc);
		enable_aux_dev(sc->kbdc);
	    } else {
		VLOG(3, (LOG_DEBUG, "psmintr: discard a byte (%d)\n",
			 sc->syncerrors));
		pb->inputbytes--;
		bcopy(&pb->ipacket[1], &pb->ipacket[0], pb->inputbytes);
	    }
	    continue;
	}

	/*
	 * We have what appears to be a valid packet.
	 * Reset the error counters.
	 */
	sc->syncerrors = 0;

	/*
	 * Drop even good packets if they occur within a timeout
	 * period of a sync error.  This allows the detection of
	 * a change in the mouse's packet mode without exposing
	 * erratic mouse behavior to the user.  Some KVMs forget
	 * enhanced mouse modes during switch events.
	 */
	if (!timeelapsed(&sc->lastinputerr, psmerrsecs, psmerrusecs, &now)) {
		pb->inputbytes = 0;
		continue;
	}

	/*
	 * Now that we're out of the validation period, reset
	 * the packet error count.
	 */
	sc->pkterrors = 0;

	sc->cmdcount++;
	if (++sc->pqueue_end >= PSM_PACKETQUEUE)
		sc->pqueue_end = 0;
	/*
	 * If we've filled the queue then call the softintr ourselves,
	 * otherwise schedule the interrupt for later.
	 */
	if (!timeelapsed(&sc->lastsoftintr, psmsecs, psmusecs, &now) ||
	    (sc->pqueue_end == sc->pqueue_start)) {
    		if ((sc->state & PSM_SOFTARMED) != 0) {
			sc->state &= ~PSM_SOFTARMED;
			untimeout(psmsoftintr, arg, sc->softcallout);
		}
		psmsoftintr(arg);
	} else if ((sc->state & PSM_SOFTARMED) == 0) {
		sc->state |= PSM_SOFTARMED;
		sc->softcallout = timeout(psmsoftintr, arg,
		    psmhz < 1 ? 1 : (hz/psmhz));
	}
    }
}

static void
psmsoftintr(void *arg)
{
    register struct psm_softc *sc = arg;
    int l;
    int s;
    packetbuf_t *pb;

    getmicrouptime(&sc->lastsoftintr);

    s = spltty();

    do {
	
	pb = &sc->pqueue[sc->pqueue_start];

        /* queue data */
        if (sc->queue.count + pb->inputbytes < sizeof(sc->queue.buf)) {
	    l = imin(pb->inputbytes, sizeof(sc->queue.buf) - sc->queue.tail);
	    bcopy(&pb->ipacket[0], &sc->queue.buf[sc->queue.tail], l);
	    if (pb->inputbytes > l)
	        bcopy(&pb->ipacket[l], &sc->queue.buf[0], pb->inputbytes - l);
            sc->queue.tail = 
		(sc->queue.tail + pb->inputbytes) % sizeof(sc->queue.buf);
            sc->queue.count += pb->inputbytes;
	}
        pb->inputbytes = 0;

	if (++sc->pqueue_start >= PSM_PACKETQUEUE)
		sc->pqueue_start = 0;
    } while (sc->pqueue_start != sc->pqueue_end);

    if (sc->state & PSM_ASLP) {
        sc->state &= ~PSM_ASLP;
        wakeup( sc);
    }
    selwakeuppri(&sc->rsel, PZERO);
    sc->state &= ~PSM_SOFTARMED;
    splx(s);
}

static int
psmpoll(struct cdev *dev, int events, struct thread *td)
{
    struct psm_softc *sc = PSM_SOFTC(PSM_UNIT(dev));
    int s;
    int revents = 0;

    /* Return true if a mouse event available */
    s = spltty();
    if (events & (POLLIN | POLLRDNORM)) {
	if (sc->queue.count > 0)
	    revents |= events & (POLLIN | POLLRDNORM);
	else
	    selrecord(td, &sc->rsel);
    }
    splx(s);

    return (revents);
}

/*
 * Return true if 'now' is earlier than (start + (secs.usecs)).
 * Now may be NULL and the function will fetch the current time from
 * getmicrouptime(), or a cached 'now' can be passed in.
 * All values should be numbers derived from getmicrouptime().
 */
static int
timeelapsed(start, secs, usecs, now)
	const struct timeval *start, *now;
	int secs, usecs;
{
	struct timeval snow, tv;

	/* if there is no 'now' passed in, the get it as a convience. */
	if (now == NULL) {
		getmicrouptime(&snow);
		now = &snow;
	}
	
	tv.tv_sec = secs;
	tv.tv_usec = usecs;
	timevaladd(&tv, start);
	return (timevalcmp(&tv, now, <));
}

static int
psmresume(device_t dev)
{
    struct psm_softc *sc = device_get_softc(dev);
    int unit = device_get_unit(dev);
    int err;

    VLOG(2, (LOG_NOTICE, "psm%d: system resume hook called.\n", unit));

    if (!(sc->config & PSM_CONFIG_HOOKRESUME))
	return (0);

    err = reinitialize(sc, sc->config & PSM_CONFIG_INITAFTERSUSPEND);

    if ((sc->state & PSM_ASLP) && !(sc->state & PSM_VALID)) {
	/* 
	 * Release the blocked process; it must be notified that the device
	 * cannot be accessed anymore.
	 */
        sc->state &= ~PSM_ASLP;
        wakeup(sc);
    }

    VLOG(2, (LOG_DEBUG, "psm%d: system resume hook exiting.\n", unit));

    return (err);
}

DRIVER_MODULE(psm, atkbdc, psm_driver, psm_devclass, 0, 0);
